/*
 * Copyright (C) 1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>	/* for malloc(), realloc() */
#include <string.h>	/* for memcpy() */

#include "compiler.h"


void *access2invoker(int access) {
  if (access & ACC_ABSTRACT) {
#ifdef RUNTIME_DEBUG
    printf("acc2inv: invokeAbstractMethod\n");  fflush(stdout);
#endif
    return (void *)sym_invokeAbstractMethod;
  }
  else if (access & ACC_NATIVE) {
#ifdef RUNTIME_DEBUG
    printf("acc2inv: invokeLazyNativeMethod\n");  fflush(stdout);
#endif
    return (void *)sym_invokeLazyNativeMethod;
  }
  else {
    if (access & ACC_SYNCHRONIZED) {
#ifdef RUNTIME_DEBUG
      printf("acc2inv: invokeSynchronizedJavaMethod\n");  fflush(stdout);
#endif
      return (void *)sym_invokeSynchronizedJavaMethod;
    }
    else {
#ifdef RUNTIME_DEBUG
      printf("acc2inv: invokeJavaMethod\n");  fflush(stdout);
#endif
      return (void *)sym_invokeJavaMethod;
    }
  }
}


void showCompilerContext(CompilerContext *cc) {
  printf("compiler context: 0x%x\n", (int)cc);
  fflush(stdout);

  if (cc) {
    struct methodblock *mb = cc->mb;

    printf("  ExecEnv: 0x%08x\n", (int)cc->ee);
    printf("  method: %s#%s %s (0x%08x)\n",
	(mb ? cbName(mb->fb.clazz) : "(null)"),
	(mb ? mb->fb.name : "(null)"),
	(mb ? mb->fb.signature : "(null)"), (int)mb);
    printf("  buffer: 0x%08x\n", (int)cc->buffer);
    printf("  bufp  : 0x%08x offset:0x%x(%d)\n",
	(int)cc->bufp, cc->bufp - cc->buffer, cc->bufp - cc->buffer);
    printf("  lastbufp: offset:0x%x(%d)\n",
	cc->lastbufp_off, cc->lastbufp_off);
    printf("  pctable: 0x%08x\n", cc->pctable);
    printf("  jptable: 0x%08x\n", cc->jptable);
  }
  fflush(stdout);
}


CompilerContext *newCompilerContext(struct methodblock *mb) {
  CompilerContext *cc;

#ifdef COMPILE_DEBUG
  printf("newCompilerContext(): mb is 0x%08x(%s)\n",
		(int)mb, (mb ? mb->fb.name : "(null)"));
  fflush(stdout);
#endif
  cc = (CompilerContext *)sysMalloc(sizeof(CompilerContext));
  if (!cc) {
#ifdef COMPILE_DEBUG
    printf("newCompilerContext(): cannot allocate memory.\n");
    fflush(stdout);
#endif
    return NULL;
  }

  cc->ee = EE();

  cc->mb = mb;
  cc->buffer = (unsigned char *)sysMalloc(DEFAULT_BUF_SIZE);
  cc->buf_size = DEFAULT_BUF_SIZE;

  cc->pctablesize = DEFAULT_PCTABLE_SIZE;
  cc->pctable = (pcentry *)sysMalloc(sizeof(pcentry) * DEFAULT_PCTABLE_SIZE);

  cc->jptablesize = DEFAULT_JPTABLE_SIZE;
  cc->jptable = (jpentry *)sysMalloc(sizeof(jpentry) * DEFAULT_JPTABLE_SIZE);

  resetCompilerContext(cc);

  return cc;
}


void freeCompilerContext(CompilerContext *cc) {
#ifdef COMPILE_DEBUG
  printf("freeCompilerContext(): cc is 0x%08x\n", (int)cc);
  fflush(stdout);
#endif
  if (cc) {
#ifdef COMPILE_DEBUG
    printf("  buffer: 0x%08x\n  pctable: 0x%08x\n  jptable: 0x%08x\n",
	(int)cc->buffer, (int)cc->pctable, (int)cc->jptable);
    fflush(stdout);
#endif
    if (cc->buffer)  sysFree(cc->buffer);
    if (cc->pctable)  sysFree(cc->pctable);
    if (cc->jptable)  sysFree(cc->jptable);
    sysFree(cc);
#ifdef COMPILE_DEBUG
    fflush(stdout);
    fflush(stderr);
#endif
  }
#ifdef COMPILE_DEBUG
  printf("freeCompilerContext() done.\n");
  fflush(stdout);
#endif
}


void resetCompilerContext(CompilerContext *cc) {
  cc->bufp = cc->buffer;  cc->lastbufp_off = 0;

  cc->pctablelen = 0;
  cc->jptablelen = 0;

  cc->exc_handler_nativeoff = -1;
  cc->finish_return_nativeoff = -1;
}


void writeToBuffer(CompilerContext *cc, void *ptr, size_t len) {
#ifdef BUF_DEBUG
  printf("writeToBuffer(cc, 0x%x, %d) called.\n", (int)ptr, (int)len);
  printf("  bufp 0x%x(offset:%d)\n", (int)cc->bufp, cc->bufp - cc->buffer);
  fflush(stdout);
#endif
  ensureBufferSize(cc, len);
  memcpy(cc->bufp, ptr, len);

  cc->lastbufp_off = cc->bufp - cc->buffer;
  cc->bufp += len;
#ifdef BUF_DEBUG
  printf("writeToBuffer() done.\n");
  fflush(stdout);
#endif
}


void cancelOnBuffer(CompilerContext *cc, size_t len) {
#ifdef BUF_DEBUG
  printf("cancelOnBuffer(cc, %d) called.\n", (int)len);
  fflush(stdout);
#endif

  cc->bufp -= len;
}


void ensureBufferSize(CompilerContext *cc, size_t req) {
  int buf_offset = cc->bufp - cc->buffer;

  if (req > (cc->buf_size - buf_offset)) {

#ifdef BUF_DEBUG
    printf("ensureBufferSize(): now extend the space.\n");
    printf("  buffer 0x%x, bufp 0x%x, offset 0x%x(%d)\n",
		(int)cc->buffer, (int)cc->bufp, buf_offset, buf_offset);
    fflush(stdout);
#endif
#ifdef BUF_DEBUG
    printf("  buf_size: %d -> ", cc->buf_size);
#endif
    do {
      cc->buf_size <<= 1;
    } while (req > (cc->buf_size - buf_offset));
#ifdef BUF_DEBUG
    printf("%d.\n", cc->buf_size);
    fflush(stdout);
#endif

    cc->buffer = (unsigned char *)sysRealloc(cc->buffer, cc->buf_size);
    cc->bufp = cc->buffer + buf_offset;
  }
}


/* PC table */

void pctableAdd(CompilerContext *cc,
	int opcode, int state, unsigned int byteoff, unsigned int nativeoff) {
  pcentry *entryp;

  if (cc->pctablelen >= cc->pctablesize) {	/* extend table size */
    do { cc->pctablesize <<= 1; } while (cc->pctablelen >= cc->pctablesize);
    cc->pctable = (pcentry *)sysRealloc(cc->pctable,
				sizeof(pcentry) * cc->pctablesize);
  }

  entryp = cc->pctable + cc->pctablelen;
  entryp->opcode = opcode;
  entryp->state = state;
  entryp->byteoff = byteoff;
  entryp->nativeoff = nativeoff;

  (cc->pctablelen)++;
}


void pctableInsert(CompilerContext *cc, int index,
	int opcode, int state, unsigned int byteoff, unsigned int nativeoff) {
  pcentry *entryp;

  if (cc->pctablelen >= cc->pctablesize) {	/* extend table size */
    do { cc->pctablesize <<= 1; } while (cc->pctablelen >= cc->pctablesize);
    cc->pctable = (pcentry *)sysRealloc(cc->pctable,
				sizeof(pcentry) * cc->pctablesize);
  }

  entryp = cc->pctable + index;
  memmove(entryp + 1, entryp, sizeof(pcentry) * (cc->pctablelen - index));

  entryp->opcode = opcode;
  entryp->state = state;
  entryp->byteoff = byteoff;
  entryp->nativeoff = nativeoff;

  (cc->pctablelen)++;
}


void pctableDelete(CompilerContext *cc, int index) {
  pcentry *entryp;

  (cc->pctablelen)--;

  entryp = cc->pctable + index;
  memmove(entryp, entryp + 1, sizeof(pcentry) * (cc->pctablelen - index));
}


pcentry *pctableGet(CompilerContext *cc, int byteoff) {
  int l = 0;
  int h = cc->pctablelen;
  int index;
  pcentry *entryp;

  while (1) {
    index = l + ((h - l) >> 1);

    entryp = cc->pctable + index;
    if (entryp->byteoff == byteoff) {	/* found */
      while (index-- >= 0) {
	if ((cc->pctable + index)->byteoff == byteoff)
	  entryp--;
	else
	  break;
      }
      return entryp;
    }

    if (l == h)  return (unsigned int)0;	/* not found */

    if (entryp->byteoff < byteoff)
      l = index + 1;
    else
      h = index;
  }
}


/* Jump instruction table */

void jptableAdd(CompilerContext *cc,
	unsigned int tgtoff, unsigned int argoff) {
  jpentry *entryp;

  if (cc->jptablelen >= cc->jptablesize) {	/* extend table size */
#ifdef COMPILE_DEBUG
    printf("jptableAdd(): extending jump table.\n");  fflush(stdout);
#endif
    do {
      cc->jptablesize <<= 1;
    } while (cc->jptablelen >= cc->jptablesize);

    cc->jptable = (jpentry *)sysRealloc(cc->jptable,
				sizeof(jpentry) * cc->jptablesize);
#ifdef COMPILE_DEBUG
    printf("jptableAdd() extending done: 0x%x\n", (int)(cc->jptable));
    fflush(stdout);
#endif
  }

  entryp = cc->jptable + cc->jptablelen;
  entryp->tgtoff = tgtoff;
  entryp->argoff = argoff;

  (cc->jptablelen)++;
}
